function [ K, z, y, relerr ] = em_isvm_multiclass( M, xtrain, xtest, ytrain, lambda, s, c, nu, l, T, T0, maxK )
%input:
%   M: integer, # of classes
%   ytrain: integer, n0 * 1, from 1 to M
%   the other parameters please see em_isvm.m

    rng(0);

    relerr = 0;
    lastloss = 1;
    
    % Initialization
    n0 = size(xtrain, 1);
    n1 = size(xtest, 1);
    n = n0 + n1;
    d = size(xtrain, 2);
    x = [xtrain; xtest];
    ext_x = [x ones(n, 1)];
    
    iytrain = zeros(n0, M);
    for m = 1:M
        iytrain(:, m) = (ytrain == m * ones(n0, 1));
    end
    
    % latent model variables
    K = 1;
    z = ones(n, 1);
    iz = zeros(n, maxK);
    iz(:, 1) = ones(n, 1);    
    mu = zeros(maxK, d);
    mu(1, :) = mean(x);    
    invomega = ones(n0, M);
    eta = zeros(maxK, M, d+1);
    zeta = zeros(n0, maxK, M);
    
    % prior variance matrix for eta
    priorL = 1/nu^2 * eye(d+1, d+1);
    priorL(d+1, d+1) = 0;
    
    pc_inf = zeros(n, maxK);
    pc_learn = zeros(n, maxK);
    
    % eta_star and their corresponding confidence
    pstar = zeros(n0, 1);
    for i = 1:n0
        eta_star = min(c*nu^2, l/norm(x(i,:))^2) * ext_x(i, :);
        zeta_star = l - ext_x(i, :) * eta_star';
        pstar(i) = 2*c * max(0, zeta_star) + norm(eta_star(1:d))^2/(2*nu^2);
    end
  
    for t = 1:T
        
        % update eta
        eta0 = eta;
        for k = 1:K
            ind_k = (iz(1:n0, k) == ones(n0, 1));
            if (sum(ind_k) == 0)
                for m = 1:M
                    eta(k, m, :) = zeros(1, d+1);
                end
                continue;
            end
            confidence = ext_x(1:n0, :) * squeeze(eta0(k, :, :))'; % n0 x M
            
            for m = 1:M
                ind_m = (iytrain(1:n0, m) == ones(n0, 1));
                
                kappa = (ind_m * 2 - ones(n0, 1)) .* ind_k;  %n0 x 1: +1: z_i = k, y_i = m; -1: z_i = k, y_i \neq m; 0: z_i\neq k
                
                t_confidence = (ones(n0, M) - iytrain) + confidence; % n0 x M
                t_confidence(:, m) = -1e30 * ones(n0, 1); % erase the m-th class
                [xi, mstar] = max(t_confidence, [], 2); % xi: n0 x 1
                xi = xi - (ones(n0, 1) - iytrain(:, m));
                
                % Lambda
                weighted_ind = ind_k .* invomega(:, m); % n0 x 1
                alt_ext_x = ext_x(1:n0, :) .* weighted_ind(:, ones(1, d+1)); % n0 x d+1
                Lambda = (priorL + c^2 * ext_x(1:n0, :)' * alt_ext_x)^(-1);
                
                % Mu
                weighted_ind = ind_k .* kappa .* (ones(n0, 1) + c * kappa .* xi .* invomega(:, m)); % n0 x 1
                eta(k, m, :) = c * weighted_ind' * ext_x(1:n0, :) * Lambda;
            end
        end
        
        % compute margins
        zeta(:, 1:K, 1:M) = zeros(n0, K, M);
        for k = 1:K
            for m = 1:M
                ind_m = iytrain(:, m);
                zeta(:, k, m) = (ones(n0, 1) - ind_m) + ext_x(1:n0, :) * squeeze(eta(k, m, :)) - sum((ext_x(1:n0, :) * squeeze(eta(k, :, :))') .* iytrain, 2);
            end
        end
        
        % predict the labels
        confidence = zeros(n, M);
        for k = 1:K
            ind_k = iz(:, k);
            confidence = confidence + ind_k(:, ones(1, M)) .* (ext_x * squeeze(eta(k, :, :))');
        end
        [maxvalue, y] = max(confidence, [], 2);
        
        % compute the pc matrix
        pc_inf(:, 1:K) = zeros(n, K);
        pc_learn(:, 1:K) = zeros(n, K);
        for k = 1:K
            % inference part
            mu0 = mu(k, :); %1 x d
            divmat = x - mu0(ones(1, n), :); % n by d, each row (x-mu0)
            divdist = sum(divmat .^ 2, 2); % n by 1
            pc_inf(:, k) = pc_inf(:, k) + s * divdist;
            % learning part
            kzeta = squeeze(zeta(:, k, :)); % n0 x M
            pc_learn(1:n0, k) = 2*c * max(kzeta, [], 2);
        end
        
        % compute the loss function, and output it
        % term 1: regularization
        A = 0;
        for k = 1:K
            for m = 1:M
                A = A + norm(squeeze(eta(k, m, 1:d)))^2 / (2*nu^2);
            end
        end
        % term 2 & 3: inference and learning
        B = 0;
        C = 0;
        for k = 1:K
            indicators = iz(:, k);
            B = B + indicators' * pc_inf(:, k);
            C = C + indicators(1:n0)' * pc_learn(1:n0, k);
        end
        % term 4: AIC model complexity
        D = lambda * K;
        
        relerr = abs((A+B+C+D) - lastloss) / lastloss;
        lastloss = A+B+C+D;
        
        fprintf('Iter %d: K = %d, loss = %f, tr acc. = %f, inf/learn ratio = %f, relerr = %f\n', ...
            t, K, A+B+C+D, mean(y(1:n0) == ytrain), B/(A+C), relerr);
        
        % update omega (inv)
        rzeta = zeros(n0, M);
        for k = 1:K
            ind_k = iz(1:n0, k);
            if (sum(ind_k) == 0)
                continue;
            end
            confidence = ext_x(1:n0, :) * squeeze(eta(k, :, :))'; % n0 x M
            for m = 1:M
                ind_m = iytrain(1:n0, m);
                kappa = (ind_m * 2 - ones(n0, 1)) .* ind_k;  %n0 x 1: +1: z_i = k, y_i = m; -1: z_i = k, y_i \neq m; 0: z_i\neq k
                t_confidence = (ones(n0, M) - iytrain) + confidence; % n0 x M
                t_confidence(:, m) = -1e30 * ones(n0, 1); % erase the m-th class
                [xi, mstar] = max(t_confidence, [], 2); % xi: n0 x 1
                xi = xi - (ones(n0, 1) - iytrain(:, m)); 
                
                rzeta(:, m) = rzeta(:, m) + ind_k .* (kappa .* xi - kappa .* (ext_x(1:n0, :) * squeeze(eta(k, m, :))));
            end
        end
        invomega = ones(n0, M) ./ (c * abs(rzeta) + 1e-4 * ones(n0, M));
        
        % update z
         if (K < maxK-1)
            pc_inf(:, K+1) = lambda * ones(n, 1);
            pc_learn(:, K+1) = zeros(n, 1);
            pc_learn(1:n0, K+1) = pstar;
        else
            pc_inf(:, K+1) = 1e100 * ones(n, 1);
            pc_learn(:, K+1) = 1e100 * ones(n, 1);
        end
        
        z0 = z;
        [minvalue, z] = min(pc_inf(:, 1:K+1) + pc_learn(:, 1:K+1), [], 2); % z: n x 1
        if ((mean(z == z0) == 1) && (t > T0))
            return;
        end
        
        K = max(z);
        for k = 1:K
            iz(:, k) = (z == k * ones(n, 1));
        end
        
        % update mu
        for k = 1:K
            indicators = iz(:, k); % n x 1
            if (sum(indicators) == 0)
                mu(k, :) = zeros(1, d);
            else
                mu(k, :) = sum(x .* indicators(:, ones(1, d))) / sum(indicators);
            end
        end       
        
        % !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        %if (K + 1 >= maxK)
        %    K = 0;
        %    z = zeros(n, 1);
        %    y = zeros(n, 1);
        %    return;
        %end
        
    end

end

